export { parseRegex, regexToNfa, nfaToDfa, minDfa, State, };
type Node = {
    begin: number;
    end: number;
    type?: string;
    text?: string;
    parts?: Node[];
    sub?: Node;
};
type State = {
    key?: string;
    items?: State[];
    symbols?: string[];
    type?: string;
    edges?: [string, State][];
    id?: string;
    trans?: Record<string, State>;
};
/**
 * Try parsing simple regular expression to syntax tree.
 *
 * Basic grammars:
 *   Empty: S -> ϵ
 *   Cat:   S -> S S
 *   Or:    S -> S | S
 *   Star:  S -> S *
 *   Text:  S -> [0-9a-zA-Z]
 *   S -> ( S )
 *
 * Extension:
 *   Plus:  S -> S + -> S S *
 *   Ques:  S -> S ? -> (S | ϵ)
 *
 * @param {string} text The input regular expression
 * @return {string|object} Returns a string that is an error message if failed to parse the expression,
 *                         otherwise returns an object which is the syntax tree.
 *
 * Edited from https://github.com/CyberZHG/toolbox/blob/gh-pages/js/lexical.js
 */
declare function parseRegex(text: string): string | Node;
/**
 * Convert regular expression to nondeterministic finite automaton.
 *
 * @param {string} text @see parseRegex()
 * @return {State|string}
 */
declare function regexToNfa(text: string): string | State;
/**
 * Convert nondeterministic finite automaton to deterministic finite automaton.
 *
 * @param {State} nfa @see regexToNfa(), the function assumes that the given NFA is valid.
 * @return {State} dfa Returns the first element of the DFA.
 */
declare function nfaToDfa(nfa: State): State;
/**
 * Convert the DFA to its minimum form using Hopcroft's algorithm.
 *
 * @param {State} dfa @see nfaToDfa(), the function assumes that the given DFA is valid.
 * @return {State} dfa Returns the first element of the minimum DFA.
 */
declare function minDfa(dfa: State): State;
