type NodeType = {
    begin: number;
    end: number;
    type?: string;
    parts?: TempNode[];
};
type TempNode = {
    begin: number;
    end: number;
    type?: string;
    sub?: NodeType;
    parts?: TempNode[];
    text?: string;
};
type StateType = {
    type: string;
    edges: [string, StateType][];
    id?: number;
};
/**
 * Try parsing simple regular expression to syntax tree.
 *
 * Basic grammars:
 *   Empty: S -> ϵ
 *   Cat:   S -> S S
 *   Or:    S -> S | S
 *   Star:  S -> S *
 *   Text:  S -> [0-9a-zA-Z]
 *   S -> ( S )
 *
 * Extension:
 *   Plus:  S -> S + -> S S *
 *   Ques:  S -> S ? -> (S | ϵ)
 *
 * @param {string} text The input regular expression
 * @return {string | object} Returns a string that is an error message if failed to parse the expression,
 *                         otherwise returns an object which is the syntax tree.
 *
 * Edited from https://github.com/CyberZHG/toolbox/blob/gh-pages/js/lexical.js
 */
declare function parseRegex(text: string): string | NodeType;
/**
 * Convert regular expression to nondeterministic finite automaton.
 *
 * @param {string} text @see parseRegex()
 * @return {object | string}
 */
declare function regexToNfa(text: string): string | {
    type: string;
    edges: never[];
};
/**
 * Convert nondeterministic finite automaton to deterministic finite automaton.
 *
 * @param {object} nfa @see regexToNfa(), the function assumes that the given NFA is valid.
 * @return {object} dfa Returns the first element of the DFA.
 */
declare function nfaToDfa(nfa: StateType): {
    key: string;
    items: any[];
    symbols: any[];
    type: string;
    edges: never[];
    trans: {};
};
/**
 * Convert the DFA to its minimum form using Hopcroft's algorithm.
 *
 * @param {object} dfa @see nfaToDfa(), the function assumes that the given DFA is valid.
 * @return {object} dfa Returns the first element of the minimum DFA.
 */
declare function minDfa(dfa: any): {
    id: string;
    key: any;
    items: never[];
    symbols: never[];
    type: any;
    edges: never[];
    trans: {};
};
